@@ -1,11 +1,14 @@
+import warnings
 import numpy as np
 
 from abc import ABCMeta, abstractmethod
 from scipy.sparse import csr_matrix
 from scipy.sparse import issparse
 
+from small_text.base import LABEL_UNLABELED
 from small_text.data.exceptions import UnsupportedOperationException
 from small_text.data.sampling import stratified_sampling, balanced_sampling
+from small_text.utils.annotations import experimental
 from small_text.utils.labels import get_flattened_unique_labels
 
 
@@ -21,6 +24,14 @@ def check_dataset_and_labels(x, y):
                          f'x.shape = {x.shape}, y.shape= {y.shape} ### {type(x)} / {type(y)}')
 
 
+def check_target_labels(target_labels):
+    if target_labels is None:
+        warnings.warn('Passing target_labels=None is discouraged as it can lead to '
+                      'unintended results in combination with indexing and cloning. '
+                      'Moreover, explicit target labels might be required in the '
+                      'next major version.')
+
+
 def get_updated_target_labels(is_multi_label, y, target_labels):
 
     if is_multi_label:
@@ -185,9 +196,18 @@ class SklearnDatasetView(DatasetView):
         else:
             y = np.copy(self.y)
 
+        dataset = self
+        while hasattr(dataset, 'dataset'):
+            dataset = dataset.dataset
+
+        if dataset.track_target_labels:
+            target_labels = None
+        else:
+            target_labels = np.copy(dataset._target_labels)
+
         return SklearnDataset(x,
                               y,
-                              target_labels=np.copy(self.target_labels))
+                              target_labels=target_labels)
 
     def __getitem__(self, item):
         return self.obj_class(self, item)
@@ -219,11 +239,12 @@ class SklearnDataset(Dataset):
         ----------
         x : numpy.ndarray or scipy.sparse.csr_matrix
             Dense or sparse feature matrix.
-        y : numpy.ndarray[int]
+        y : numpy.ndarray[int] or scipy.sparse.csr_matrix
             List of labels where each label belongs to the features of the respective row.
         target_labels : numpy.ndarray[int] or None, default=None
             List of possible labels. Will be inferred from `y` if `None` is passed."""
         check_dataset_and_labels(x, y)
+        check_target_labels(target_labels)
 
         self._x = x
         self._y = y
@@ -238,7 +259,21 @@ class SklearnDataset(Dataset):
             self._infer_target_labels()
 
     def _infer_target_labels(self):
-        self.target_labels = get_flattened_unique_labels(self)
+        if self.x.shape[0] == 0:
+            self.target_labels = np.array([0])
+        else:
+            unique_labels = np.unique(self._get_flattened_labels())
+            if unique_labels.shape[0] > 0:
+                max_label_id = unique_labels.max()
+                self.target_labels = np.arange(max_label_id+1)
+            else:
+                self.target_labels = np.array([0])
+
+    def _get_flattened_labels(self):
+        if self.multi_label:
+            return self.y.indices
+        else:
+            return self.y[np.argwhere(self.y > LABEL_UNLABELED)]
 
     @property
     def x(self):
@@ -265,7 +300,16 @@ class SklearnDataset(Dataset):
         check_dataset_and_labels(self.x, y)
         self._y = y
 
-        self.target_labels = get_updated_target_labels(self.is_multi_label, y, self.target_labels)
+        if self.track_target_labels:
+            self.target_labels = get_updated_target_labels(self.is_multi_label, y, self.target_labels)
+        else:
+            max_label_id = np.max(y)
+            max_target_labels_id = self.target_labels.max()
+            if max_label_id > max_target_labels_id:
+                raise ValueError(f'Error while assigning new labels to dataset: '
+                                 f'Encountered label with id {max_label_id} which is outside of '
+                                 f'the configured set of target labels (whose maximum label is '
+                                 f'is {max_target_labels_id}) [track_target_labels=False]')
 
     @property
     def is_multi_label(self):
@@ -284,6 +328,10 @@ class SklearnDataset(Dataset):
 
     @target_labels.setter
     def target_labels(self, target_labels):
+        encountered_labels = get_flattened_unique_labels(self)
+        if np.setdiff1d(encountered_labels, target_labels).shape[0] > 0:
+            raise ValueError('Cannot remove existing labels from target_labels as long as they '
+                             'still exists in the data. Create a new dataset instead.')
         self._target_labels = target_labels
 
     def clone(self):
@@ -297,7 +345,56 @@ class SklearnDataset(Dataset):
         else:
             y = np.copy(self._y)
 
-        return SklearnDataset(x, y, target_labels=np.copy(self._target_labels))
+        if self.track_target_labels:
+            target_labels = None
+        else:
+            target_labels = np.copy(self._target_labels)
+
+        return SklearnDataset(x, y, target_labels=target_labels)
+
+    @classmethod
+    @experimental
+    def from_arrays(cls, texts, y, vectorizer, target_labels=None, train=True):
+        """Constructs a new SklearnDataset from the given text and label arrays.
+
+        Parameters
+        ----------
+        texts : list of str or np.ndarray[str]
+            List of text documents.
+        y : np.ndarray[int] or scipy.sparse.csr_matrix
+            List of labels where each label belongs to the features of the respective row.
+            Depending on the type of `y` the resulting dataset will be single-label (`np.ndarray`)
+            or multi-label (`scipy.sparse.csr_matrix`).
+        vectorizer : object
+            A scikit-learn vectorizer which is used to construct the feature matrix.
+        target_labels : numpy.ndarray[int] or None, default=None
+            List of possible labels. Will be directly passed to the datset constructor.
+        train : bool
+            If `True` fits the vectorizer and transforms the data, otherwise just transforms the
+            data.
+
+        Returns
+        -------
+        dataset : SklearnDataset
+            A dataset constructed from the given texts and labels.
+
+
+        .. seealso::
+           `scikit-learn docs: Vectorizer API reference
+           <https://scikit-learn.org/stable/modules/classes.html
+           #module-sklearn.feature_extraction.text>`_
+
+        .. warning::
+           This functionality is still experimental and may be subject to change.
+
+        .. versionadded:: 1.1.0
+        """
+        if train:
+            x = vectorizer.fit_transform(texts)
+        else:
+            x = vectorizer.transform(texts)
+
+        return SklearnDataset(x, y, target_labels=target_labels)
 
     def __getitem__(self, item):
         return SklearnDatasetView(self, item)
