@@ -15,9 +15,11 @@ from small_text.integrations.pytorch.models.kimcnn import KimCNN
 from small_text.utils.classification import empty_result, get_splits
 from small_text.utils.context import build_pbar_context
 from small_text.utils.data import check_training_data, list_length
+from small_text.utils.annotations import early_stopping_deprecation_warning
 from small_text.utils.labels import csr_to_list, get_num_labels
 from small_text.utils.datetime import format_timedelta
 from small_text.utils.logging import verbosity_logger, VERBOSITY_MORE_VERBOSE
+from small_text.utils.system import get_tmp_dir_base
 
 
 try:
@@ -26,17 +28,17 @@ try:
     from torch.optim import Adadelta
 
     from small_text.integrations.pytorch.classifiers.base import (
-        check_optimizer_and_scheduler_config
+        _check_optimizer_and_scheduler_config
     )
     from small_text.integrations.pytorch.datasets import PytorchTextClassificationDataset
-    from small_text.integrations.pytorch.model_selection import Metric, PytorchModelSelection
     from small_text.integrations.pytorch.utils.data import dataloader
 except ImportError:
     raise PytorchNotFoundError('Could not import pytorch')
 
 
-def kimcnn_collate_fn(batch, enc=None, max_seq_len=60, padding_idx=0, filter_padding=0):
-
+def kimcnn_collate_fn(batch, multi_label=None, num_classes=None, use_sample_weights=False,
+                      max_seq_len=60, padding_idx=0, filter_padding=0):
+    # TODO: torch.no_grad()?
     def prepare_tensor(t):
         t_sub = t[:max_seq_len-2*filter_padding]
         return torch.cat([t_sub.new_zeros(filter_padding) + padding_idx,
@@ -45,15 +47,24 @@ def kimcnn_collate_fn(batch, enc=None, max_seq_len=60, padding_idx=0, filter_pad
                           t_sub.new_zeros(filter_padding) + padding_idx],
                          0)
 
-    if enc is not None:
-        labels = [entry[PytorchTextClassificationDataset.INDEX_LABEL] for entry in batch]
-        multi_hot = enc.transform(labels)
+    if multi_label:
+        multi_hot = [[0 if i not in set(entry[PytorchTextClassificationDataset.INDEX_LABEL]) else 1
+                      for i in range(num_classes)]
+                     for entry in batch]
         label = torch.tensor(multi_hot, dtype=float)
     else:
-        label = torch.tensor([entry[PytorchTextClassificationDataset.INDEX_LABEL] for entry in batch])
-    text = torch.stack([prepare_tensor(t) for t, _ in batch], 0)
+        label = torch.tensor([entry[PytorchTextClassificationDataset.INDEX_LABEL]
+                              for entry in batch])
+    text = torch.stack([prepare_tensor(entry[PytorchTextClassificationDataset.INDEX_TEXT])
+                        for entry in batch], 0)
 
-    return text, label
+    if use_sample_weights:
+        weights = torch.tensor([entry[-1]
+                                for entry in batch])
+    else:
+        weights = torch.ones(text.size(0), device=text.device)
+
+    return text, label, weights
 
 
 class KimCNNEmbeddingMixin(EmbeddingMixin):
@@ -75,7 +86,7 @@ class KimCNNEmbeddingMixin(EmbeddingMixin):
         tensors = []
         proba = []
         with build_pbar_context(pbar, tqdm_kwargs={'total': list_length(data_set)}) as pbar:
-            for text, _ in dataset_iter:
+            for text, *_ in dataset_iter:
                 batch_len = text.size(0)
                 text = text.to(self.device, non_blocking=True)
 
@@ -154,10 +165,55 @@ class KimCNNClassifier(KimCNNEmbeddingMixin, PytorchClassifier):
     def __init__(self, num_classes, multi_label=False, embedding_matrix=None, device=None,
                  num_epochs=10, mini_batch_size=25, lr=0.001, max_seq_len=60, out_channels=100,
                  filter_padding=0, dropout=0.5, validation_set_size=0.1, padding_idx=0,
-                 kernel_heights=[3, 4, 5], early_stopping=5, early_stopping_acc=0.98,
+                 kernel_heights=[3, 4, 5], early_stopping=5, early_stopping_acc=-1,
                  class_weight=None, verbosity=VERBOSITY_MORE_VERBOSE):
+        """
+        num_classes : int
+            Number of classes.
+        multi_label : bool, default=False
+            If `False`, the classes are mutually exclusive, i.e. the prediction step results in
+            exactly one predicted label per instance.
+        embedding_matrix : torch.FloatTensor
+            A tensor of embeddings in the shape of (vocab_size, embedding_size).
+        device : str or torch.device, default=None
+            Torch device on which the computation will be performed.
+        num_epochs : int, default=10
+            Epochs to train.
+        mini_batch_size : int, default=12
+            Size of mini batches during training.
+        lr : float, default=2e-5
+            Learning rate.
+        max_seq_len : int
+            Maximum sequence length.
+        out_channels : int
+            Number of output channels.
+        filter_padding : int
+            Size of the padding to add before and after the sequence before applying the filters.
+        dropout : float
+            Dropout probability for the final layer in KimCNN.
+        validation_set_size : float, default=0.1
+            The size of the validation set as a fraction of the training set.
+        padding_idx : int
+            Index of the padding token (as given by the `vocab`).
+        kernel_heights : list of int
+            Kernel sizes.
+        early_stopping : int
+            Number of epochs with no improvement in validation loss until early stopping
+            is triggered.
 
+             .. deprecated:: 1.1.0
+               Use the `early_stopping` kwarg in `fit()` instead.
+        early_stopping_acc : float
+            Accuracy threshold in the interval (0, 1] which triggers early stopping.
+
+            .. deprecated:: 1.1.0
+               Use the `early_stopping` kwarg in `fit()` instead.
+        class_weight : 'balanced' or None, default=None
+            If 'balanced', then the loss function is weighted inversely proportional to the
+            label distribution to the current train set.
+        """
         super().__init__(multi_label=multi_label, device=device, mini_batch_size=mini_batch_size)
+        early_stopping_deprecation_warning(early_stopping, early_stopping_acc)
 
         with verbosity_logger():
             self.logger = logging.getLogger(__name__)
@@ -194,7 +250,8 @@ class KimCNNClassifier(KimCNNEmbeddingMixin, PytorchClassifier):
         self.model_selection = None
         self.enc_ = None
 
-    def fit(self, train_set, validation_set=None, optimizer=None, scheduler=None):
+    def fit(self, train_set, validation_set=None, weights=None, early_stopping=None,
+            model_selection=None, optimizer=None, scheduler=None):
         """Trains the model using the given train set.
 
         Parameters
@@ -203,8 +260,14 @@ class KimCNNClassifier(KimCNNEmbeddingMixin, PytorchClassifier):
             The dataset used for training the model.
         validation_set : PytorchTextClassificationDataset
             A validation set used for validation during training, or `None`. If `None`, the fit
-            operation will split apart a subset of the trainset as a validation set, whose size
+            operation will split apart a subset of the train set as a validation set, whose size
             is set by `self.validation_set_size`.
+        weights : np.ndarray[np.float32] or None, default=None
+            Sample weights or None.
+        early_stopping : EarlyStoppingHandler or 'none'
+            A strategy for early stopping. Passing 'none' disables early stopping.
+        model_selection : ModelSelectionHandler or 'none'
+            A model selection handler. Passing 'none' disables model selection.
         optimizer : torch.optim.optimizer.Optimizer
             A pytorch optimizer.
         scheduler :torch.optim._LRScheduler
@@ -215,10 +278,32 @@ class KimCNNClassifier(KimCNNEmbeddingMixin, PytorchClassifier):
         self : KimCNNClassifier
             Returns the current classifier with a fitted model.
         """
-        check_training_data(train_set, validation_set)
+        check_training_data(train_set, validation_set, weights=weights)
 
-        sub_train, sub_valid = get_splits(train_set, validation_set, multi_label=self.multi_label,
-                                          validation_set_size=self.validation_set_size)
+        if weights is not None:
+            sub_train, sub_valid, sub_train_weights = get_splits(
+                train_set,
+                validation_set,
+                weights=weights,
+                multi_label=self.multi_label,
+                validation_set_size=self.validation_set_size
+            )
+        else:
+            sub_train, sub_valid = get_splits(
+                train_set,
+                validation_set,
+                multi_label=self.multi_label,
+                validation_set_size=self.validation_set_size
+            )
+            sub_train_weights = None
+
+        early_stopping = self._get_default_early_stopping(
+            early_stopping,
+            self.early_stopping,
+            self.early_stopping_acc,
+            1,
+            kwarg_no_improvement_name='early_stopping')
+        model_selection = self._get_default_model_selection(model_selection)
 
         fit_optimizer = optimizer if optimizer is not None else self.optimizer
         fit_scheduler = scheduler if scheduler is not None else self.scheduler
@@ -229,26 +314,24 @@ class KimCNNClassifier(KimCNNEmbeddingMixin, PytorchClassifier):
             self.enc_ = self.enc_.fit(labels)
 
         self.class_weights_ = self.initialize_class_weights(sub_train)
-        self.criterion = self.get_default_criterion()
+        self.criterion = self._get_default_criterion(self.class_weights_,
+                                                     use_sample_weights=weights is not None)
 
-        return self._fit_main(sub_train, sub_valid, fit_optimizer, fit_scheduler)
+        return self._fit_main(sub_train, sub_valid, sub_train_weights, early_stopping,
+                              model_selection, fit_optimizer, fit_scheduler)
 
-    def _fit_main(self, sub_train, sub_valid, optimizer, scheduler):
+    def _fit_main(self, sub_train, sub_valid, weights, early_stopping, model_selection,
+                  optimizer, scheduler):
         if self.model is None:
             encountered_num_classes = get_num_labels(sub_train.y)
 
             if self.num_classes is None:
                 self.num_classes = encountered_num_classes
 
-            if self.num_classes != encountered_num_classes:
-                raise ValueError('Conflicting information about the number of classes: '
-                                 'expected: {}, encountered: {}'.format(self.num_classes,
-                                                                        encountered_num_classes))
-
             self.initialize_kimcnn_model(sub_train)
 
-        check_optimizer_and_scheduler_config(optimizer, scheduler)
-        scheduler = scheduler if scheduler is None else None
+        _check_optimizer_and_scheduler_config(optimizer, scheduler)
+        scheduler = scheduler if scheduler is not None else None
 
         optimizer, scheduler = self._get_optimizer_and_scheduler(optimizer,
                                                                  scheduler,
@@ -256,13 +339,12 @@ class KimCNNClassifier(KimCNNEmbeddingMixin, PytorchClassifier):
                                                                  sub_train)
 
         self.model = self.model.to(self.device)
-        with tempfile.TemporaryDirectory() as tmp_dir:
-            res = self._train(sub_train, sub_valid, tmp_dir, optimizer, scheduler)
+        with tempfile.TemporaryDirectory(dir=get_tmp_dir_base()) as tmp_dir:
+            self._train(sub_train, sub_valid, weights, early_stopping, model_selection,
+                        optimizer, scheduler, tmp_dir)
+            self._perform_model_selection(optimizer, model_selection)
 
-            model_path, _ = self.model_selection.select_best()
-            self.model.load_state_dict(torch.load(model_path))
-
-        return res
+        return self
 
     def initialize_kimcnn_model(self, sub_train):
         vocab_size = len(sub_train.vocab)
@@ -279,66 +361,63 @@ class KimCNNClassifier(KimCNNEmbeddingMixin, PytorchClassifier):
         params = [param for param in self.model.parameters() if param.requires_grad]
         return params, Adadelta(params, lr=base_lr, eps=1e-8)
 
-    def _train(self, sub_train, sub_valid, tmp_dir, optimizer, scheduler):
-
-        min_loss = float('inf')
-        no_loss_reduction = 0
-
-        metrics = [Metric('valid_loss', True), Metric('valid_acc', False),
-                   Metric('train_loss', True), Metric('train_acc', False)]
-        self.model_selection = PytorchModelSelection(tmp_dir, metrics=metrics)
+    def _train(self, sub_train, sub_valid, weights, early_stopping, model_selection, optimizer,
+               scheduler, tmp_dir):
 
+        stop = False
         for epoch in range(self.num_epochs):
-            start_time = datetime.datetime.now()
+            if not stop:
+                start_time = datetime.datetime.now()
 
-            self.model.train()
-            train_loss, train_acc = self._train_func(sub_train, optimizer, scheduler)
+                self.model.train()
+                train_loss, train_acc = self._train_func(sub_train,
+                                                         weights,
+                                                         optimizer,
+                                                         scheduler)
 
-            self.model.eval()
-            valid_loss, valid_acc = self.validate(sub_valid)
-            self.model_selection.add_model(self.model, epoch+1, valid_acc=valid_acc,
-                                           valid_loss=valid_loss, train_acc=train_acc,
-                                           train_loss=train_loss)
+                self.model.eval()
+                valid_loss, valid_acc = self.validate(sub_valid)
 
-            timedelta = datetime.datetime.now() - start_time
+                timedelta = datetime.datetime.now() - start_time
 
-            self.logger.info(f'Epoch: {epoch+1} | {format_timedelta(timedelta)}\n'
-                             f'\tTrain Set Size: {len(sub_train)}\n'
-                             f'\tLoss: {train_loss:.4f}(train)\t|\tAcc: {train_acc * 100:.1f}% (train)\n'
-                             f'\tLoss: {valid_loss:.4f}(valid)\t|\tAcc: {valid_acc * 100:.1f}% (valid)',
-                             verbosity=VERBOSITY_MORE_VERBOSE)
+                self.logger.info(f'Epoch: {epoch + 1} | {format_timedelta(timedelta)}\n'
+                                 f'\tTrain Set Size: {len(sub_train)}\n'
+                                 f'\tLoss: {train_loss:.4f}(train)\t|\tAcc: {train_acc * 100:.1f}% (train)\n'
+                                 f'\tLoss: {valid_loss:.4f}(valid)\t|\tAcc: {valid_acc * 100:.1f}% (valid)',
+                                 verbosity=VERBOSITY_MORE_VERBOSE)
 
-            if self.early_stopping > 0:
-                if valid_loss < min_loss:
-                    no_loss_reduction = 0
-                    min_loss = valid_loss
-                else:
-                    no_loss_reduction += 1
-
-                    if no_loss_reduction >= self.early_stopping:
-                        print('\nEarly stopping after %s epochs' % (epoch+1))
-                        return self
-
-            if self.early_stopping_acc > 0:
-                if train_acc > self.early_stopping_acc:
-                    print('\nEarly stopping due to high train acc: %s' % (train_acc))
-                    return self
+                measured_values = {
+                    'train_loss': train_loss,
+                    'train_acc': train_acc,
+                    'val_loss': valid_loss,
+                    'val_acc': valid_acc
+                }
+                stop = early_stopping.check_early_stop(epoch + 1, measured_values)
+                self._save_model(optimizer, model_selection, f'{epoch}-0',
+                                 train_acc, train_loss, valid_acc, valid_loss, stop, tmp_dir)
 
         return self
 
-    def _create_collate_fn(self):
-        return partial(kimcnn_collate_fn, enc=self.enc_, padding_idx=self.padding_idx,
-                       max_seq_len=self.max_seq_len, filter_padding=self.filter_padding)
+    def _create_collate_fn(self, use_sample_weights=False):
+        return partial(kimcnn_collate_fn, multi_label=self.multi_label,
+                       num_classes=self.num_classes, use_sample_weights=use_sample_weights,
+                       padding_idx=self.padding_idx, max_seq_len=self.max_seq_len,
+                       filter_padding=self.filter_padding)
 
-    def _train_func(self, sub_train_, optimizer, scheduler):
+    def _train_func(self, sub_train_, weights, optimizer, scheduler):
 
         train_loss = 0.
         train_acc = 0.
 
-        train_iter = dataloader(sub_train_.data, self.mini_batch_size, self._create_collate_fn())
+        data = sub_train_.data
+        if weights is not None:
+            data = [d + (weights[i],) for i, d in enumerate(data)]
 
-        for i, (text, cls) in enumerate(train_iter):
-            loss, acc = self._train_single_batch(text, cls, optimizer)
+        train_iter = dataloader(data, self.mini_batch_size,
+                                self._create_collate_fn(use_sample_weights=weights is not None))
+
+        for i, (text, cls, weight) in enumerate(train_iter):
+            loss, acc = self._train_single_batch(text, cls, weight, optimizer)
             scheduler.step()
 
             train_loss += loss
@@ -346,14 +425,14 @@ class KimCNNClassifier(KimCNNEmbeddingMixin, PytorchClassifier):
 
         return train_loss / len(sub_train_), train_acc / len(sub_train_)
 
-    def _train_single_batch(self, text, cls, optimizer):
+    def _train_single_batch(self, text, cls, weight, optimizer):
 
         train_loss = 0.
         train_acc = 0.
 
         optimizer.zero_grad()
 
-        text, cls = text.to(self.device), cls.to(self.device)
+        text, cls, weight = text.to(self.device), cls.to(self.device), weight.to(self.device)
         output = self.model(text)
 
         with torch.no_grad():
@@ -362,12 +441,12 @@ class KimCNNClassifier(KimCNNEmbeddingMixin, PytorchClassifier):
             else:
                 target = cls
         loss = self.criterion(output, target)
+        loss = loss * weight
+        loss = loss.mean()
 
         loss.backward()
 
-        with torch.no_grad():
-            torch.nn.utils.clip_grad_norm_(self.model.parameters(), 3)
-            self.model.fc.weight.div_(torch.norm(self.model.fc.weight, dim=1, keepdim=True))
+        torch.nn.utils.clip_grad_norm_(self.model.parameters(), 3)
 
         optimizer.step()
 
@@ -401,8 +480,8 @@ class KimCNNClassifier(KimCNNEmbeddingMixin, PytorchClassifier):
         valid_iter = dataloader(validation_set.data, self.mini_batch_size, self._create_collate_fn(),
                                 train=False)
 
-        for x, cls in valid_iter:
-            x, cls = x.to(self.device), cls.to(self.device)
+        for x, cls, weight, *_ in valid_iter:
+            x, cls, weight = x.to(self.device), cls.to(self.device), weight.to(self.device)
 
             with torch.no_grad():
                 if self.num_classes == 2:
@@ -412,6 +491,9 @@ class KimCNNClassifier(KimCNNEmbeddingMixin, PytorchClassifier):
 
                 output = self.model(x)
                 loss = self.criterion(output, target)
+                loss = loss * weight
+                loss = loss.mean()
+
                 valid_loss += loss.item()
 
                 acc += self.sum_up_accuracy_(output, cls)
@@ -440,20 +522,20 @@ class KimCNNClassifier(KimCNNEmbeddingMixin, PytorchClassifier):
         """
         return super().predict(data_set, return_proba=return_proba)
 
-    def predict_proba(self, test_set):
-        if len(test_set) == 0:
+    def predict_proba(self, data_set):
+        if len(data_set) == 0:
             return empty_result(self.multi_label, self.num_classes, return_prediction=False,
                                 return_proba=True)
 
         self.model.eval()
-        test_iter = dataloader(test_set.data, self.mini_batch_size, self._create_collate_fn(),
+        test_iter = dataloader(data_set.data, self.mini_batch_size, self._create_collate_fn(),
                                train=False)
 
         predictions = []
         logits_transform = torch.sigmoid if self.multi_label else partial(F.softmax, dim=1)
 
         with torch.no_grad():
-            for text, _ in test_iter:
+            for text, *_ in test_iter:
                 text = text.to(self.device)
 
                 predictions += logits_transform(self.model.forward(text)).to('cpu').tolist()
