@@ -13,7 +13,7 @@ try:
     import torch
     import torch.nn.functional as F  # noqa: N812
 
-    from small_text.integrations.pytorch.utils.misc import assert_layer_exists, default_tensor_type
+    from small_text.integrations.pytorch.utils.misc import _assert_layer_exists
     from small_text.integrations.pytorch.utils.data import dataloader
 except ImportError:
     raise PytorchNotFoundError('Could not import pytorch')
@@ -61,8 +61,8 @@ class ExpectedGradientLength(QueryStrategy):
         pbar_context = build_pbar_context('tqdm', tqdm_kwargs={'total': list_length(dataset)})
 
         offset = 0
-        with default_tensor_type(torch.FloatTensor), pbar_context as pbar:
-            for i, (dataset, _) in enumerate(dataset_iter):
+        with pbar_context as pbar:
+            for i, (dataset, *_) in enumerate(dataset_iter):
                 self.compute_gradient_lengths(clf, criterion, gradient_lengths, offset, dataset)
 
                 batch_len = dataset.size(0)
@@ -180,7 +180,7 @@ class ExpectedGradientLengthMaxWord(ExpectedGradientLength):
         if clf.model is None:
             raise ValueError('Initial model must be trained!')
 
-        assert_layer_exists(clf.model, self.layer_name)
+        _assert_layer_exists(clf.model, self.layer_name)
         modules = dict(clf.model.named_modules())
         if not isinstance(modules[self.layer_name], torch.nn.Embedding):
             raise ValueError(f'Given parameter (layer_name={self.layer_name}) '
@@ -261,7 +261,7 @@ class ExpectedGradientLengthLayer(ExpectedGradientLength):
 
     def compute_gradient_length(self, clf, sm, gradients, j, k):
 
-        assert_layer_exists(clf.model, self.layer_name)
+        _assert_layer_exists(clf.model, self.layer_name)
 
         modules = dict({name: module for name, module in clf.model.named_modules()})
         params = [param.grad.flatten() for param in modules[self.layer_name].parameters()
