@@ -6,9 +6,6 @@ import warnings
 import numpy as np
 
 from functools import partial
-from pathlib import Path
-
-from sklearn.preprocessing import MultiLabelBinarizer
 
 from small_text.classifiers.classification import EmbeddingMixin
 from small_text.integrations.pytorch.exceptions import PytorchNotFoundError
@@ -16,7 +13,11 @@ from small_text.utils.classification import empty_result, get_splits
 from small_text.utils.context import build_pbar_context
 from small_text.utils.data import check_training_data, list_length
 from small_text.utils.datetime import format_timedelta
-from small_text.utils.labels import csr_to_list, get_num_labels
+from small_text.utils.annotations import (
+    early_stopping_deprecation_warning,
+    model_selection_deprecation_warning
+)
+from small_text.utils.labels import get_num_labels
 from small_text.utils.logging import verbosity_logger, VERBOSITY_MORE_VERBOSE
 from small_text.utils.system import get_tmp_dir_base
 
@@ -29,28 +30,34 @@ try:
     from transformers import AutoConfig, AutoModelForSequenceClassification, AutoTokenizer
 
     from small_text.integrations.pytorch.classifiers.base import (
-        check_optimizer_and_scheduler_config,
+        _check_optimizer_and_scheduler_config,
         PytorchClassifier
     )
-    from small_text.integrations.pytorch.model_selection import Metric, PytorchModelSelection
     from small_text.integrations.pytorch.utils.data import dataloader
     from small_text.integrations.transformers.datasets import TransformersDataset
 except ImportError:
     raise PytorchNotFoundError('Could not import pytorch')
 
 
-def transformers_collate_fn(batch, enc=None):
+def transformers_collate_fn(batch, multi_label=None, num_classes=None, use_sample_weights=False):
     with torch.no_grad():
         text = torch.cat([entry[TransformersDataset.INDEX_TEXT] for entry in batch], dim=0)
         masks = torch.cat([entry[TransformersDataset.INDEX_MASK] for entry in batch], dim=0)
-        if enc is not None:
-            labels = [entry[TransformersDataset.INDEX_LABEL] for entry in batch]
-            multi_hot = enc.transform(labels)
+        if multi_label:
+            multi_hot = [[0 if i not in set(entry[TransformersDataset.INDEX_LABEL]) else 1
+                         for i in range(num_classes)]
+                         for entry in batch]
             label = torch.tensor(multi_hot, dtype=float)
         else:
             label = torch.tensor([entry[TransformersDataset.INDEX_LABEL] for entry in batch])
 
-    return text, masks, label
+    if use_sample_weights:
+        weights = torch.tensor([entry[-1]
+                                for entry in batch])
+    else:
+        weights = torch.ones(text.size(0), device=text.device)
+
+    return text, masks, label, weights
 
 
 class FineTuningArguments(object):
@@ -178,7 +185,7 @@ class TransformerBasedEmbeddingMixin(EmbeddingMixin):
 
     def _create_embeddings(self, tensors, batch, embedding_method='avg', hidden_layer_index=-1):
 
-        text, masks, _ = batch
+        text, masks, *_ = batch
         text = text.to(self.device, non_blocking=True)
         masks = masks.to(self.device, non_blocking=True)
 
@@ -226,13 +233,20 @@ class TransformerBasedClassification(TransformerBasedEmbeddingMixin, PytorchClas
         mini_batch_size : int, default=12
             Size of mini batches during training.
         validation_set_size : float, default=0.1
-            The sizes of the validation as a fraction of the training set.
+            The size of the validation set as a fraction of the training set.
         validations_per_epoch : int, default=1
             Defines how of the validation set is evaluated during the training of a single epoch.
         early_stopping_no_improvement : int, default=5
-            Number of epochs of no improvement in validation loss until early stopping is triggered.
+            Number of epochs with no improvement in validation loss until early stopping
+            is triggered.
+
+            .. deprecated:: 1.1.0
+               Use the `early_stopping` kwarg in `fit()` instead.
         early_stopping_acc : float, default=-1
-            Accuracy threshold in the interval (0, 1] whose exceeding triggers early stopping.
+            Accuracy threshold in the interval (0, 1] which triggers early stopping.
+
+            .. deprecated:: 1.1.0
+               Use the `early_stopping` kwarg in `fit()` instead.
         model_selection : bool, default=True
             If True, model selects first saves the model after each epoch. At the end of the
             training step the model with the lowest validation error is selected.
@@ -248,6 +262,8 @@ class TransformerBasedClassification(TransformerBasedEmbeddingMixin, PytorchClas
             label distribution to the current train set.
         """
         super().__init__(multi_label=multi_label, device=device, mini_batch_size=mini_batch_size)
+        early_stopping_deprecation_warning(early_stopping_no_improvement, early_stopping_acc)
+        model_selection_deprecation_warning(model_selection)
 
         with verbosity_logger():
             self.logger = logging.getLogger(__name__)
@@ -282,11 +298,9 @@ class TransformerBasedClassification(TransformerBasedEmbeddingMixin, PytorchClas
         self.model = None
         self.model_selection_manager = None
 
-        self.enc_ = None
-
-    def fit(self, train_set, validation_set=None, optimizer=None, scheduler=None):
-        """
-        Trains the model using the given train set.
+    def fit(self, train_set, validation_set=None, weights=None, early_stopping=None,
+            model_selection=None, optimizer=None, scheduler=None):
+        """Trains the model using the given train set.
 
         Parameters
         ----------
@@ -296,6 +310,12 @@ class TransformerBasedClassification(TransformerBasedEmbeddingMixin, PytorchClas
             A validation set used for validation during training, or `None`. If `None`, the fit
             operation will split apart a subset of the trainset as a validation set, whose size
             is set by `self.validation_set_size`.
+        weights : np.ndarray[np.float32] or None, default=None
+            Sample weights or None.
+        early_stopping : EarlyStoppingHandler or 'none'
+            A strategy for early stopping. Passing 'none' disables early stopping.
+        model_selection : ModelSelectionHandler or 'none'
+            A model selection handler. Passing 'none' disables model selection.
         optimizer : torch.optim.optimizer.Optimizer or None, default=None
             A pytorch optimizer.
         scheduler : torch.optim._LRScheduler or None, default=None
@@ -306,45 +326,60 @@ class TransformerBasedClassification(TransformerBasedEmbeddingMixin, PytorchClas
         self : TransformerBasedClassification
             Returns the current classifier with a fitted model.
         """
-        check_training_data(train_set, validation_set)
+        check_training_data(train_set, validation_set, weights=weights)
 
         optimizer_or_scheduler_given = optimizer is not None or scheduler is not None
         if self.fine_tuning_arguments is not None and optimizer_or_scheduler_given:
             raise ValueError('When fine_tuning_arguments are provided you cannot pass '
                              'optimizer and scheduler to fit()')
 
-        sub_train, sub_valid = get_splits(train_set, validation_set, multi_label=self.multi_label,
-                                          validation_set_size=self.validation_set_size)
+        if weights is not None:
+            sub_train, sub_valid, sub_train_weights = get_splits(
+                train_set,
+                validation_set,
+                weights=weights,
+                multi_label=self.multi_label,
+                validation_set_size=self.validation_set_size
+            )
+        else:
+            sub_train, sub_valid = get_splits(
+                train_set,
+                validation_set,
+                multi_label=self.multi_label,
+                validation_set_size=self.validation_set_size
+            )
+            sub_train_weights = None
+
+        early_stopping = self._get_default_early_stopping(
+            early_stopping,
+            self.early_stopping_no_improvement,
+            self.early_stopping_acc,
+            self.validations_per_epoch,
+            kwarg_no_improvement_name='early_stopping_no_improvement')
+        model_selection = self._get_default_model_selection(model_selection)
 
         fit_optimizer = optimizer if optimizer is not None else self.optimizer
         fit_scheduler = scheduler if scheduler is not None else self.scheduler
 
-        if self.multi_label:
-            self.enc_ = MultiLabelBinarizer()
-            labels = csr_to_list(sub_train.y)
-            self.enc_ = self.enc_.fit(labels)
-
         self.class_weights_ = self.initialize_class_weights(sub_train)
-        self.criterion = self.get_default_criterion()
+        self.criterion = self._get_default_criterion(self.class_weights_,
+                                                     use_sample_weights=weights is not None)
 
-        return self._fit_main(sub_train, sub_valid, fit_optimizer, fit_scheduler)
+        return self._fit_main(sub_train, sub_valid, sub_train_weights, early_stopping,
+                              model_selection, fit_optimizer, fit_scheduler)
 
-    def _fit_main(self, sub_train, sub_valid, optimizer, scheduler):
+    def _fit_main(self, sub_train, sub_valid, weights, early_stopping, model_selection,
+                  optimizer, scheduler):
         if self.model is None:
             encountered_num_classes = get_num_labels(sub_train.y)
 
             if self.num_classes is None:
                 self.num_classes = encountered_num_classes
 
-            if self.num_classes != encountered_num_classes:
-                raise ValueError('Conflicting information about the number of classes: '
-                                 'expected: {}, encountered: {}'.format(self.num_classes,
-                                                                        encountered_num_classes))
-
             self.initialize_transformer(self.cache_dir)
 
-        check_optimizer_and_scheduler_config(optimizer, scheduler)
-        scheduler = scheduler if scheduler is None else 'linear'
+        _check_optimizer_and_scheduler_config(optimizer, scheduler)
+        scheduler = scheduler if scheduler is not None else 'linear'
 
         optimizer, scheduler = self._get_optimizer_and_scheduler(optimizer,
                                                                  scheduler,
@@ -353,10 +388,11 @@ class TransformerBasedClassification(TransformerBasedEmbeddingMixin, PytorchClas
         self.model = self.model.to(self.device)
 
         with tempfile.TemporaryDirectory(dir=get_tmp_dir_base()) as tmp_dir:
-            res = self._train(sub_train, sub_valid, tmp_dir, optimizer, scheduler)
-            self._perform_model_selection(sub_valid)
+            self._train(sub_train, sub_valid, weights, early_stopping, model_selection,
+                        optimizer, scheduler, tmp_dir)
+            self._perform_model_selection(optimizer, model_selection)
 
-        return res
+        return self
 
     def initialize_transformer(self, cache_dir):
 
@@ -390,67 +426,33 @@ class TransformerBasedClassification(TransformerBasedEmbeddingMixin, PytorchClas
 
         return params, AdamW(params, lr=base_lr, eps=1e-8)
 
-    def _train(self, sub_train, sub_valid, tmp_dir, optimizer, scheduler):
+    def _train(self, sub_train, sub_valid, weights, early_stopping, model_selection, optimizer,
+               scheduler, tmp_dir):
 
-        metrics = [Metric('valid_loss', True), Metric('valid_acc', False),
-                   Metric('train_loss', True), Metric('train_acc', False)]
-        self.model_selection_manager = PytorchModelSelection(Path(tmp_dir), metrics)
+        stop = False
+        for epoch in range(0, self.num_epochs):
+            if not stop:
+                start_time = datetime.datetime.now()
 
-        start_epoch = 0
-        self._train_loop(sub_train, sub_valid, optimizer, scheduler, start_epoch, self.num_epochs,
-                         self.model_selection_manager)
+                train_acc, train_loss, valid_acc, valid_loss, stop = self._train_loop_epoch(epoch,
+                                                                                            sub_train,
+                                                                                            sub_valid,
+                                                                                            weights,
+                                                                                            early_stopping,
+                                                                                            model_selection,
+                                                                                            optimizer,
+                                                                                            scheduler,
+                                                                                            tmp_dir)
 
-        return self
+                timedelta = datetime.datetime.now() - start_time
 
-    def _train_loop(self, sub_train, sub_valid, optimizer, scheduler, start_epoch, num_epochs,
-                    model_selection_manager):
+                self._log_epoch(epoch, timedelta, sub_train, sub_valid, train_acc, train_loss,
+                                valid_acc, valid_loss)
 
-        min_loss = float('inf')
-        no_loss_reduction = 0
+    def _train_loop_epoch(self, num_epoch, sub_train, sub_valid, weights, early_stopping,
+                          model_selection, optimizer, scheduler, tmp_dir):
 
-        stopped = False
-
-        for epoch in range(start_epoch, num_epochs):
-            start_time = datetime.datetime.now()
-
-            train_acc, train_loss, valid_acc, valid_loss = self._train_loop_epoch(sub_train,
-                                                                                  sub_valid,
-                                                                                  epoch,
-                                                                                  optimizer,
-                                                                                  scheduler)
-
-            timedelta = datetime.datetime.now() - start_time
-
-            self._log_epoch(epoch, timedelta, sub_train, sub_valid, train_acc, train_loss,
-                            valid_acc, valid_loss)
-
-            if sub_valid is not None:
-                if self.early_stopping_no_improvement > 0:
-                    if valid_loss < min_loss:
-                        no_loss_reduction = 0
-                        min_loss = valid_loss
-                    else:
-                        no_loss_reduction += 1
-
-                        if no_loss_reduction >= self.early_stopping_no_improvement:
-                            logging.info(f'Early stopping after {epoch + 1} epochs')
-                            stopped = True
-
-                if not stopped and self.early_stopping_acc > 0:
-                    if train_acc > self.early_stopping_acc:
-                        logging.info(f'Early stopping due to high train acc: {train_acc}')
-                        stopped = True
-
-                model_selection_manager.add_model(self.model, epoch + 1, valid_acc=valid_acc,
-                                                  valid_loss=valid_loss, train_acc=train_acc,
-                                                  train_loss=train_loss)
-
-            if stopped:
-                break
-
-    def _train_loop_epoch(self, sub_train, sub_valid, epoch, optimizer, scheduler):
-
-        if self.memory_fix and (epoch + 1) % self.memory_fix == 0:
+        if self.memory_fix and (num_epoch + 1) % self.memory_fix == 0:
             torch.cuda.empty_cache()
 
         self.model.train()
@@ -465,56 +467,88 @@ class TransformerBasedClassification(TransformerBasedEmbeddingMixin, PytorchClas
                 validate_every = 1
             else:
                 validate_every = int(num_batches / self.validations_per_epoch)
-
-            train_loss, train_acc, valid_loss, valid_acc = self._train_loop_process_batches(
-                sub_train,
-                optimizer,
-                scheduler,
-                validate_every=validate_every,
-                validation_set=sub_valid)
         else:
-            train_loss, train_acc = self._train_loop_process_batches(
-                sub_train,
-                optimizer,
-                scheduler)
+            validate_every = None
 
-            if sub_valid is not None:
-                valid_loss, valid_acc = self.validate(sub_valid)
+        train_loss, train_acc, valid_loss, valid_acc, stop = self._train_loop_process_batches(
+            num_epoch,
+            sub_train,
+            sub_valid,
+            weights,
+            early_stopping,
+            model_selection,
+            optimizer,
+            scheduler,
+            tmp_dir,
+            validate_every=validate_every)
 
-        return train_acc, train_loss, valid_acc, valid_loss
+        return train_acc, train_loss, valid_acc, valid_loss, stop
 
-    def _train_loop_process_batches(self, sub_train_, optimizer, scheduler, validate_every=None,
-                                    validation_set=None):
+    def _train_loop_process_batches(self, num_epoch, sub_train_, sub_valid_, weights, early_stopping,
+                                    model_selection, optimizer, scheduler, tmp_dir,
+                                    validate_every=None):
 
         train_loss = 0.
         train_acc = 0.
         valid_losses = []
         valid_accs = []
 
-        train_iter = dataloader(sub_train_.data, self.mini_batch_size, self._create_collate_fn())
+        data = sub_train_.data
+        if weights is not None:
+            data = [d + (weights[i],) for i, d in enumerate(data)]
 
-        for i, (x, masks, cls) in enumerate(train_iter):
-            loss, acc = self._train_single_batch(x, masks, cls, optimizer)
-            scheduler.step()
+        train_iter = dataloader(data, self.mini_batch_size,
+                                self._create_collate_fn(use_sample_weights=weights is not None))
 
-            train_loss += loss
-            train_acc += acc
+        stop = False
 
-            if validate_every and i % validate_every == 0:
-                valid_loss, valid_acc = self.validate(validation_set)
-                valid_losses.append(valid_loss)
-                valid_accs.append(valid_acc)
+        for i, (x, masks, cls, weight, *_) in enumerate(train_iter):
+            if not stop:
+                loss, acc = self._train_single_batch(x, masks, cls, weight, optimizer)
+                scheduler.step()
+
+                train_loss += loss
+                train_acc += acc
+
+                if validate_every and i % validate_every == 0:
+                    valid_loss, valid_acc = self.validate(sub_valid_)
+                    valid_losses.append(valid_loss)
+                    valid_accs.append(valid_acc)
+
+                    measured_values = dict({
+                        'val_loss': valid_loss,
+                        'val_acc': valid_acc
+                    })
+                    stop = stop or early_stopping.check_early_stop(num_epoch+1, measured_values)
+                    self._save_model(optimizer, model_selection, f'{num_epoch}-b{i+1}',
+                                     train_acc, train_loss, valid_acc, valid_loss, stop, tmp_dir)
 
         if validate_every:
-            return train_loss / len(sub_train_), train_acc / len(sub_train_), \
-                   np.mean(valid_losses), np.mean(valid_accs)
+            valid_loss, valid_acc = np.mean(valid_losses), np.mean(valid_accs)
+
         else:
-            return train_loss / len(sub_train_), train_acc / len(sub_train_)
+            valid_loss, valid_acc = None, None
+            if validate_every is None and sub_valid_ is not None:
+                valid_loss, valid_acc = self.validate(sub_valid_)
+        train_loss = train_loss / len(sub_train_)
+        train_acc = train_acc / len(sub_train_)
 
-    def _create_collate_fn(self):
-        return partial(transformers_collate_fn, enc=self.enc_)
+        measured_values = {
+            'train_loss': train_loss,
+            'train_acc': train_acc,
+            'val_loss': valid_loss,
+            'val_acc': valid_acc
+        }
+        stop = early_stopping.check_early_stop(num_epoch+1, measured_values)
+        self._save_model(optimizer, model_selection, f'{num_epoch}-b0',
+                         train_acc, train_loss, valid_acc, valid_loss, stop, tmp_dir)
+        return train_loss, train_acc, valid_loss, valid_acc, stop
 
-    def _train_single_batch(self, x, masks, cls, optimizer):
+    def _create_collate_fn(self, use_sample_weights=False):
+        return partial(transformers_collate_fn, multi_label=self.multi_label,
+                       num_classes=self.num_classes, use_sample_weights=use_sample_weights)
+
+    def _train_single_batch(self, x, masks, cls, weight, optimizer):
 
         train_loss = 0.
         train_acc = 0.
@@ -522,9 +556,13 @@ class TransformerBasedClassification(TransformerBasedEmbeddingMixin, PytorchClas
         optimizer.zero_grad()
 
         x, masks, cls = x.to(self.device), masks.to(self.device), cls.to(self.device)
+        weight = weight.to(self.device)
+
         outputs = self.model(x, attention_mask=masks)
 
         logits, loss = self._compute_loss(cls, outputs)
+        loss = loss * weight
+        loss = loss.mean()
 
         loss.backward()
 
@@ -550,21 +588,6 @@ class TransformerBasedClassification(TransformerBasedEmbeddingMixin, PytorchClas
 
         return logits, loss
 
-    def _perform_model_selection(self, sub_valid):
-        if sub_valid is not None:
-            if self.model_selection:
-                self._select_best_model()
-            else:
-                self._select_last_model()
-
-    def _select_best_model(self):
-        model_path, _ = self.model_selection_manager.select_best()
-        self.model.load_state_dict(torch.load(model_path))
-
-    def _select_last_model(self):
-        model_path, _ = self.model_selection_manager.select_last()
-        self.model.load_state_dict(torch.load(model_path))
-
     def _log_epoch(self, epoch, timedelta, sub_train, sub_valid, train_acc, train_loss, valid_acc, valid_loss):
         if sub_valid is not None:
             valid_loss_txt = f'\n\tLoss: {valid_loss:.4f}(valid)\t|\tAcc: {valid_acc * 100:.1f}%(valid)'
@@ -582,15 +605,19 @@ class TransformerBasedClassification(TransformerBasedEmbeddingMixin, PytorchClas
         acc = 0.
 
         self.model.eval()
-        valid_iter = dataloader(validation_set.data, self.mini_batch_size, self._create_collate_fn(),
+        valid_iter = dataloader(validation_set.data, self.mini_batch_size,
+                                self._create_collate_fn(),
                                 train=False)
 
-        for x, masks, cls in valid_iter:
+        for x, masks, cls, weight, *_ in valid_iter:
             x, masks, cls = x.to(self.device), masks.to(self.device), cls.to(self.device)
+            weight = weight.to(self.device)
 
             with torch.no_grad():
                 outputs = self.model(x, attention_mask=masks)
                 _, loss = self._compute_loss(cls, outputs)
+                loss = loss * weight
+                loss = loss.mean()
 
                 valid_loss += loss.item()
                 acc += self.sum_up_accuracy_(outputs.logits, cls)
@@ -617,20 +644,20 @@ class TransformerBasedClassification(TransformerBasedEmbeddingMixin, PytorchClas
         """
         return super().predict(data_set, return_proba=return_proba)
 
-    def predict_proba(self, test_set):
-        if len(test_set) == 0:
+    def predict_proba(self, data_set):
+        if len(data_set) == 0:
             return empty_result(self.multi_label, self.num_classes, return_prediction=False,
                                 return_proba=True)
 
         self.model.eval()
-        test_iter = dataloader(test_set.data, self.mini_batch_size, self._create_collate_fn(),
+        test_iter = dataloader(data_set.data, self.mini_batch_size, self._create_collate_fn(),
                                train=False)
 
         predictions = []
         logits_transform = torch.sigmoid if self.multi_label else partial(F.softmax, dim=1)
 
         with torch.no_grad():
-            for text, masks, _ in test_iter:
+            for text, masks, *_ in test_iter:
                 text, masks = text.to(self.device), masks.to(self.device)
                 outputs = self.model(text, attention_mask=masks)
 
